{% extends "nbdimepage.html" %}

    {% block nbdimeheader %}

    <div id="nbdime-header" class="nbdime-Merge">
      <h3>Notebook Merge</h3>
      <div id="nbdime-header-buttonrow">
        <label><input id="nbdime-hide-unchanged" type="checkbox"></input>Hide unchanged cells</label>
        <button id="nbdime-save" style="display: none">Save</button>
        <button id="nbdime-download" style="display: none">Download</button>
        <button id="nbdime-close" style="display: none">Close tool</button>
      </div>
      <div id=nbdime-header-banner>
        <span id="nbdime-header-local">Local</span>
        <span id="nbdime-header-base">Base</span>
        <span id="nbdime-header-remote">Remote</span>
      </div>
    </div> <!-- ndime-header -->

    {% endblock %}
