{% extends "nbdimepage.html" %}

    {% block nbdimeheader %}

    <div id="nbdime-header" class="nbdime-Diff">
      <h3>Notebook Diff</h3>
      Enter notebook filenames or URLs in the form below to get started.

      <form id="nbdime-diff-form" class="nbdime-forms">
        <fieldset>
          <legend>Please input filenames/URLs of notebooks to diff:</legend>
          <label>Base:
          <input id="diff-base" type="text" name="base" value="{{ config_data['base']|e }}" /></label>
          <label>Remote:
          <input id="diff-remote" type="text" name="remote" value="{{ config_data['remote']|e }}" /></label>
          <input id="nbdime-run-diff" type="submit" name="diff" value="Diff files" />
        </fieldset>
      </form> <!-- nbdime-forms -->
      <div id="nbdime-header-buttonrow">
        <label><input id="nbdime-hide-unchanged" type="checkbox"></input>Hide unchanged cells</label>
        <button id="nbdime-trust" style="display: none">Trust outputs</button>
        <button id="nbdime-close" type="checkbox" style="display: none">Close tool</button>
        <button id="nbdime-export" type="checkbox" style="display: none">Export diff</button>
      </div>
      <div id="nbdime-header-banner" class="cm-merge-2pane">
        <span id="nbdime-header-base">Base</span>
        <span id="nbdime-header-remote">Remote</span>
      </div>
    </div> <!-- ndime-header -->

    {% endblock %}
