{% extends "nbdimepage.html" %}

    {% block nbdimeheader %}

    <div id="nbdime-header" class="nbdime-Diff">
      <h3>Notebook Diff</h3>
      <div id="nbdime-header-buttonrow">
        <label><input id="nbdime-hide-unchanged" type="checkbox"></input>Hide unchanged cells</label>
        <button id="nbdime-trust" style="display: none">Trust outputs</button>
        <button id="nbdime-close" style="display: none">Close tool</button>
        <button id="nbdime-export" style="display: none">Export diff</button>
      </div>
      <div id="nbdime-header-banner" class="cm-merge-2pane">
        <span id="nbdime-header-base">Base</span>
        <span id="nbdime-header-remote">Remote</span>
      </div>
    </div> <!-- ndime-header -->

    {% endblock %}